#!/bin/sh
#
# kernelmod.sh
# Copyright (C) 2003  Richard de Vroede <richard@fenix.getmyip.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# Your copy is available at: http://www.gnu.org/licenses/gpl.html

version="0.7.1"

patchdir=$(echo $0 | rev | cut -d "/" -f2- | rev)
cd $patchdir
patchdir=`pwd`
kmoddir=$patchdir
debug=0
BSD=1

PTCH=$(which patch || exit 1)
CC=$(which cc || exit 1)
MODPRB=$(which modprobe || exit 1)
LD=$(which ld || exit 1)

k_ver=$(cat /proc/sys/kernel/osrelease)
maj_ver=$(echo $k_ver | awk -F "." '{print $1 "." $2}')
min_ver=$(echo $k_ver | cut -d "." -f3 | cut -d "-" -f1)
release=$(echo $k_ver | cut -d "-" -f2)

if [ -f /etc/redhat-release ]; then
   DISTRO="RedHat"
   DVER=$(cat /etc/redhat-release | cut -d " " -f5-)
elif [ -f /etc/SuSE-release ]; then 
   DISTRO="SuSE"
   DVER=$(cat /etc/SuSE-release | head -n1 | cut -d " " -f3)
elif [ -f /etc/debian_version ]; then 
   DISTRO="Debian"
   DVER=$(cat /etc/debian_version)
else
   DISTRO=$os
fi

if [ $debug ]; then
    KMODLOG=$kmoddir/debug.log
else
    KMODLOG=/dev/null
fi

header() {
    echo "-------------------------------------------------"
    echo "  KernelMod v$version"
    echo "  Richard de Vroede <richard@fenix.getmyip.com>"
    echo "  James Cameron <quozl@us.netrek.org>"
    echo "-------------------------------------------------"
}

findsource() {
   echo "--> Finding source for your running kernel."
   if [ "$DISTRO" = "RedHat" -a -d "/usr/src/linux-$k_ver" ]; then
      KSOURCE="/usr/src/linux-$k_ver"
   fi
   if [ "$DISTRO" = "Debian" ]; then
      if [ -d "/usr/src/kernel-source-$maj_ver.$min_ver" ]; then
         KSOURCE="/usr/src/kernel-source-$maj_ver.$min_ver"
         echo "Found $KSOURCE, this probably is what we need."
         FOUND=1
      else
         echo "/usr/src/kernel-source-$maj_ver.$min_ver not found."
         echo "Install the kernel-source belonging to your running kernel!"
         echo "apt-get install kernel-source-$maj_ver.$min_ver"
	 echo "cd /usr/src/"
	 echo "tar -xjf kernel-source-$maj_ver.$min_ver.tar.bz2"
         exit 1
      fi
   fi
   if [ ! -z "$KSOURCE" ]; then
# Found a candidate, check the version.
      TEMP1=$(cat $KSOURCE/include/linux/version.h 2>$KMODLOG | grep `uname -r`)
      if [ ! -z "TEMP1" ]; then FOUND=1; fi
   fi
   if [ -z "$FOUND" ]; then
# Not found yet.
      for lindir in `ls -1d /usr/src/linux*`; do
# Is lindir a symlink?
         TEMP=$(file $lindir | cut -d " " -f2)
         if [ "$TEMP" = "symbolic" -o "$TEMP" = "broken" ]; then
# Yes, skip it.
            echo "$lindir is a symlink. Skipping it."
         else
# No, check the version.
            TEMP1=$(cat $lindir/include/linux/version.h 2>$KMODLOG | grep `uname -r`)
            if [ -n "$TEMP1" -o "$TEMP1" != "" ]; then
# Found it. 
               KSOURCE=$lindir
               FOUND=1
            fi
         fi
      done
   fi
   if [ ! -z "$FOUND" ]; then
      echo "Found the source for your running kernel."
      echo "Location: $KSOURCE"
      return
   else
      echo "Didn't find it. Cowardly bailing out."
      exit 1
   fi
}

rollbackpatch() {
    echo "--> Rolling back patches."
    if [ -z "$KSOURCE" ]; then 
	findsource
    elif [ -z "$error" ]; then
	echo "You specified the location of the kernelsource."
	echo "I trust your judgement, so I won't verify."
    fi
    echo "-------------------------------------------------"
    for patch in `ls -1 $patchdir/$maj_ver/*-$maj_ver.$min_ver-*.patch`; do
    echo "Reverting patch $patch."
    echo "-------------------------------------------------"
        for file in `cat $patch | grep "\-\-\- " | awk -F " " '{print $2}' | sed 's/.orig//g' | cut -d "/" -f2-`; do
	    if [ -f "$KSOURCE/$file.bak" ]; then
	        echo -n "Processing file $file..."
                mv -f "$KSOURCE/$file.bak" "$KSOURCE/$file"
	        echo "Done."
            fi
        done
    done
}
version() {
cat <<EOF

   KernelMod v$version

EOF
}

usage() {
   header
cat <<EOF

   Usage: $0 [-a architecture] [-o] [-e emulation] [-h] [-k kernelsource] [-r] [-s]

   Options:
      -a    Define architecture (Default: i386)
      -e    Define emulation (Supported emulations: elf_i386 i386linux
            elf_i386_glibc21; Default: elf_i386)
      -h    Shows this help
      -k    The location of the kernelsource
      -o    Use OpenSSL licensed module instead of BSD licensed module
            (you need ppp < 2.4.2 to use this)
      -r    Revert the patches
      -s    Build for SMP (Default: No SMP)
      -v    Show version information

EOF
}
while getopts ":a:e:hk:orsv" opt; do
   case $opt in
      a) arch=$OPTARG ;;
      e) emulation=$OPTARG ;;
      h) usage; exit ;;
      k) KSOURCE=$OPTARG ;;
      o) patchdir="$patchdir/openssl"; BSD=0 ;;
      r) header; rollbackpatch; exit ;;
      s) smp="1" ;;
      v) version; exit ;;
     \?) echo "Wrong option(s)"; usage; exit 1 ;;
   esac
done

shift $(($OPTIND-1))

if [ $BSD -eq 1 ]; then
   groups="ppp_mppe"
   group_ppp_mppe="ppp_mppe_compress.o sha1.o arcfour.o"
   excludes=""
else
   groups=""
   excludes="rc4_enc.o rc4_skey.o sha1dgst.o"
fi

header
echo "--> Summary."
os=$(uname)
echo "OS:             $os"

echo "Distribution:   $DISTRO $DVER"

echo "Kernel version: $k_ver"

if [ -z "$arch" ]; then 
    arch=$(uname -m)
fi
echo "Architecture:   $arch"

# Supported emulations: elf_i386 i386linux elf_i386_glibc21
if [ -z "$emulation" ]; then 
    emulation="elf_i386"
fi
echo "Emulation:      $emulation"

echo -n "SMP:            "
if [ -n "$smp" ]; then echo "Yes"; else echo "No"; fi

if [ $BSD -eq 1 ]; then
   license="BSD without advertisement clause"
else
   license="OpenSSL"
fi
echo "License:        $license"

echo "-------------------------------------------------"
echo "--> Preparing system."
mkdir -p $kmoddir/build
echo "Clearing out builddir $kmoddir/build"
rm -rf $kmoddir/build/*

if [ -z "$KSOURCE" ]; then 
   findsource
else
   echo "You specified the location of the kernelsource."
   if [ ! -d "$KSOURCE" ]; then
      echo "It's not there!!!!"
      exit 1
   fi
   echo "I trust your judgement, so I won't verify."
   echo "I will set the correct kernelversion though."
   VERSION=$(echo $maj_ver | cut -d . -f1)
   PATCHLEVEL=$(echo $maj_ver | cut -d . -f2)
   SUBLEVEL=$min_ver
   EXTRAVERSION="-$release"
   LINES=$(wc -l $KSOURCE/Makefile | awk -F " " '{print $1}')
   LINES=$(($LINES-4))
   tail -n $LINES $KSOURCE/Makefile > /tmp/Makefile
   echo -e "VERSION = $VERSION\nPATCHLEVEL = $PATCHLEVEL\nSUBLEVEL = $SUBLEVEL\nEXTRAVERSION = $EXTRAVERSION\n" > $KSOURCE/Makefile
   cat /tmp/Makefile >> $KSOURCE/Makefile
   make -s symlinks -C $KSOURCE
   make -s include/linux/version.h -C $KSOURCE
fi

echo "--> Making sure the kernelsource is configured."
if [ ! -f "$KSOURCE/.config" ]; then 
   cd $KSOURCE
if [ "$DISTRO" = "RedHat" ]; then
   # Does not exist in standard kernel. Have to make a patch for the Makefile.
   make oldconfig_nonint 2>$KMODLOG 1>$KMODLOG || echo "Failed!"
else
   make oldconfig || echo "Failed!"
fi
   cd $patchdir
   echo "Done."
else
   echo "Found $KSOURCE/.config, so it probably is configured."
fi
echo "--> Looking for kernel-headers and a good modversions.h"
#       echo "Updating dependencies."
#       make dep 2>$KMODLOG 1>$KMODLOG || echo "Failed!"
# This breaks building of modules on RedHat, Debian and maybe others. 
# Needed to make modversions.h, so let's do something with kernel-headers
# for Debian.
if [ $DISTRO = "Debian" ]; then
   if [ -d "/usr/src/kernel-headers-$k_ver" ]; then
      KHEADERS="/usr/src/kernel-headers-$k_ver"
      echo "Found $KHEADERS, this probably is what we need."
   else
      echo "/usr/src/kernel-headers-$k_ver not found."
      echo "Install the kernel-headers belonging to your running kernel!"
      echo "apt-get install kernel-headers-$k_ver"
      exit 1
   fi
fi
if [ ! -z "$KHEADERS" ]; then
   KINCLUDE=$(echo "$KHEADERS/include")
else
   KINCLUDE=$(echo "$KSOURCE/include")
fi

# RedHat 6.2, kernel 2.2.x
if [ -f "$KSOURCE/include/config/modversions.h" -a ! -f $KSOURCE/include/linux/modversions.h ]; then
   ln -s $KSOURCE/include/config/modversions.h $KSOURCE/include/linux/modversions.h
fi

echo "--> Done."

echo "-------------------------------------------------"
echo "--> Locating patches."

if [ -d "$KSOURCE" ]; then
    if [ -d "$patchdir/$maj_ver" ]; then
	DEFAULTS=$(ls -1 $patchdir/$maj_ver/*-default-*.patch)
        if [ -n "$DEFAULTS" ]; then
	    echo "Found patches for $maj_ver."
	    echo "Checking for specific patches."
	    for patch in `ls -1 $patchdir/$maj_ver/*-default-*.patch`; do
		pre=$(echo $patch | rev | cut -d "/" -f1 | rev | awk -F "-default-" '{print $1}')
		post=$(echo $patch | rev | cut -d "/" -f1 | rev  | awk -F "-default-" '{print $2}' | sed 's/\.patch//g')
		if [ ! -f "$patchdir/$maj_ver/$pre-$maj_ver.$min_ver-$post.patch" ]; then
		    CURDIR=`pwd`
		    cd $patchdir/$maj_ver
		    ln -s $pre-default-$post.patch $pre-$maj_ver.$min_ver-$post.patch
		    cd $CURDIR
		fi
	    done
	fi
	PATCHES=$(ls -1 $patchdir/$maj_ver/*-$maj_ver.$min_ver-*.patch)
        if [ -n "$PATCHES" ]; then
	    echo "Found patches for $maj_ver.$min_ver"
	    echo "-------------------------------------------------"
	    echo "--> Patches & sources"
	    for patch in `ls -1 $patchdir/$maj_ver/*-$maj_ver.$min_ver-*.patch`; do
		echo "Applying patch $patch"
		$PTCH -p1 -N -b -z .bak -d "$KSOURCE" -i $patch || error=1
                if [ "$error" = "1" ]; then
                    echo "Found problems in module(s) from patch $patch."
                    rollbackpatch
                    exit 1
		fi
		for hfile in `cat $patch | grep "\-\-\- " | grep "\.h" | awk -F " " '{print $2}' | sed 's/.orig//g' | cut -d "/" -f2-`; do
		    cp -f "$KSOURCE/$hfile" $kmoddir/build
		    if [ ! -z "$KHEADERS" ]; then
		       LOC=$(echo $hfile | rev | cut -d "/" -f2- | rev)
		       if [ -d "$KHEADERS/$LOC" ]; then
		          echo "Copying $hfile to $KHEADERS"
		          cp -f "$KSOURCE/$hfile" $KHEADERS/$hfile
		       fi
		    fi
		done
		for cfile in `cat $patch | grep "\-\-\- " | grep "\.c" | awk -F " " '{print $2}' | sed 's/.orig//g' | cut -d "/" -f2-`; do
		    cp -f "$KSOURCE/$cfile" $kmoddir/build
		done
		if [ -d "$kmoddir/src" ]; then
		   echo "Copying extra sources to $KSOURCE"
		   for srcdir in `ls -1 $kmoddir/src | grep -v README | grep -v CVS`; do
		      for srcfile in `cat $kmoddir/src/$srcdir/meta.kmod | grep -v "^#" | grep "^index\:"`; do
			 srcfilebase=$(echo $srcfile | cut -d ":" -f2- | rev | cut -d "/" -f1 | rev)
			 srcfilename=$(echo $srcfile | cut -d ":" -f2-)
			 echo "$srcfilebase --> $KSOURCE/$srcfilename"
			 cp -f "$kmoddir/src/$srcdir/$srcfilebase" $KSOURCE/$srcfilename
		      done
		      echo "Copying extra sources to $kmoddir/build/"
		      cp -f "$kmoddir/src/$srcdir/"*.[ch] $kmoddir/build/
		   done
		fi
		if [ "$error" = "1" ]; then
		    echo "Found problems in patch $patch."
		    rollbackpatch
		    exit 1
		fi
	    done
	    echo "-------------------------------------------------"
            if [ ! -d "/lib/modules/$k_ver" ]; then
               echo "Hmmm. Can't find the /lib/modules/$k_ver directory."
               echo "Maybe it's somewhere else."
               echo "You'll have to copy the built modules there yourself."
               echo "You can find them in $kmoddir/build/"
	       exit 1
            fi
	    for patch in `ls -1 $patchdir/$maj_ver/*-$maj_ver.$min_ver-*.patch`; do
		cd $kmoddir/build
		# [mod1]for cfile in `cat $patch | grep "\-\-\- " | grep "\.c" | awk -F " " '{print $2}' | sed 's/.orig//g' | cut -d "/" -f2-`; do
		for cfile in `ls -1 *.c`; do
    	    	    # [mod1]basename=$(echo $cfile | rev | cut -d "/" -f1 | sed 's/^c.//g' | rev)
    	    	    basename=$(echo $cfile | rev | cut -d "/" -f1 | sed 's/^c.//g' | rev)
		    TEMP=$(echo $excludes | grep "$basename")
		    if [ -z "$TEMP" ]; then
			echo "Building module $basename.o"
			CFLAGS=" -D__KERNEL__"
			CFLAGS=$(echo "$CFLAGS -D__KERNEL_MODULE_$arch=1")
			if [ -n "$smp" ]; then CFLAGS=$(echo "$CFLAGS -D__BOOT_KERNEL_SMP=1"); fi
			CFLAGS=$(echo "$CFLAGS -Wall -Wstrict-prototypes -Wno-trigraphs -O2 -fomit-frame-pointer -fno-strict-aliasing -fno-common -Wno-unused -pipe")
			if [ "$maj_ver" != "2.2" ]; then
			    CFLAGS=$(echo "$CFLAGS -mpreferred-stack-boundary=2")
			fi
			CFLAGS=$(echo "$CFLAGS -I$KINCLUDE")
			CFLAGS=$(echo "$CFLAGS -include $KINCLUDE/linux/modversions.h")
			CFLAGS=$(echo "$CFLAGS -march=$arch -DMODULE -DMODVERSIONS")
			CFLAGS=$(echo "$CFLAGS -DKBUILD_BASENAME=$basename -DEXPORT_SYMTAB")
		        cc $CFLAGS -c $basename.c || error=1
	    	    fi
		done
		for group in `echo $groups`; do
			mods=$(echo "\$group_$group")
			basename=$group
                	echo "Building module $basename.o"
			$LD -m $emulation -A $arch -r -o $basename.o `eval "echo $mods"`
			rm -f `eval "echo $mods"`
                done
		echo "-------------------------------------------------"
		if [ "$error" = "1" ]; then
		    echo "Found problems in module(s) from patch $patch."
		    rollbackpatch
		    exit 1
		fi
		for destfile in `cat $patchdir/src/$srcdir/meta.kmod | grep -v "^#" | grep "^dest\:"`; do
		    basename=$(echo $destfile | rev | cut -d "/" -f1 | sed 's/^o.//g' | rev)
		    TEMP=`eval "echo $mods"`
		    TEMP=$(echo $TEMP | grep $basename)
		    TEMP=$(echo $TEMP | cut -d "." -f1)
                    ddir=$(echo $destfile | cut -d ":" -f2- | rev | cut -d "/" -f2- | rev)
        	    if [ "$maj_ver" = "2.4" ]; then ddir=$(echo "kernel/$ddir"); fi
	    	    echo "Installing module $basename.o in /lib/modules/$k_ver/$ddir/"
		    mv -f $kmoddir/build/$basename.o /lib/modules/$k_ver/$ddir/
		done
		cd $patchdir
	    done
	else
	    echo "Didn't find any patches for $maj_ver.$min_ver. Bailing out."
	    exit 1
	fi
    else
	echo "Didn't find the directory with patches for $maj_ver. Bailing out."
	exit 1
    fi
else
    echo "You don't seem to have the kernel-source files in $KSOURCE. Bailing out."
    exit 1
fi

for module in `lsmod | grep ppp | awk '{print $1}'`; do rmmod -r $module 2>$KMODLOG; done
echo "Updating module dependencies"
depmod -a
modprobe ppp_mppe 2>$KMODLOG 1>$KMODLOG
TEMP=$(lsmod | grep ppp_mppe)
if [ -z "$TEMP" ]; then 
    echo "WARNING: ppp_mppe module not loaded! Must be something wrong:"
    modprobe ppp_mppe
    error=1
fi

if [ "$error" = "1" ]; then
    echo "There has been an error. Leaving buildstuff in $kmoddir/build"
    rollbackpatch
    exit 1
else
    echo "Everything seems OK. Removing buildstuff in $kmoddir/build"
    rm -rf $kmoddir/build/*
fi

if [ "$BSD" = "0" ]; then cd ..; fi

if [ "$error" = "1" ]; then exit 1; fi

if [ "$DISTRO" = "RedHat" ]; then
    echo "-------------------------------------------------"
    echo "You are running RedHat, do you want me to make an RPM for you?"
    echo "Note: the modules are already installed, this RPM is for future use"
    echo "      or distribution."
    echo -n "Make an RPM? (yes/no) > "
    read reply
    if [ "$reply" = "yes" -o "$reply" = "y" -o "$reply" = "Y" ]; then
        cp -f confmod.sh /usr/src/redhat/SOURCES/
        if [ "$maj_ver" = "2.2" ]; then
            rpmbuild -bb kernel-mppe-2.2.spec --target "$arch--$os" --define "VERSION $maj_ver.$min_ver" --define "RELEASE $release" --define "BSD $BSD" --define "LICENSE $license" --quiet 1>$KMODLOG 2>$KMODLOG || exit 1
        else
            rpmbuild -bb kernel-mppe.spec --target "$arch--$os" --define "VERSION $maj_ver.$min_ver" --define "RELEASE $release" --define "BSD $BSD" --define "LICENSE $license" --quiet 1>$KMODLOG 2>$KMODLOG || exit 1
        fi
	rpm -U /usr/src/redhat/RPMS/$arch/kernel-mppe-$k_ver.$arch.rpm --justdb --quiet
        echo "Done. RPM: /usr/src/redhat/RPMS/$arch/kernel-mppe-$k_ver.$arch.rpm"
    fi
fi

exit 0
